// bot_fire.c

#include "bot.h"
#include "bot_move.h"
#include "bot_fire.h"

void B_FireBFG (mobj_t* actor);

bool  dont_damage_player;
bool  hit_player;

// spawn bot's rocket
bool B_SpawnRocket ( mobj_t*       source,
                     mobj_t*       dest,
                     mobjtype_t    type )
{
    mobj_t*     th;
    angle_t     an;
    int         dist;
    int         i;
    fixed_t     momx, momy, momz;

    th = P_SpawnMobj (source->x,
                      source->y,
                      source->z + 4*8*FRACUNIT, type);

	if (th->info->seesound)
		S_Sound (th, CHAN_VOICE, th->info->seesound, 1, ATTN_NORM);

    th->target = source;        // where it came from

    dist = P_AproxDistance (dest->x - source->x, dest->y - source->y);
    dist = dist / th->info->speed;
    dist = P_AproxDistance (dest->x + dest->momx*dist - source->x,
                            dest->y + dest->momy*dist - source->y);
    dist = dist / th->info->speed;

    // pridict player position
    an = R_PointToAngle2 (source->x, source->y,
        dest->x + dest->momx*dist, dest->y + dest->momy*dist);

    if (dist < 1)
     dist = 1;

    th->angle = an;
    an >>= ANGLETOFINESHIFT;
    th->momx = FixedMul (th->info->speed, finecosine[an]);
    th->momy = FixedMul (th->info->speed, finesine[an]);
    th->momz = (dest->z - source->z) / dist;

    momx = th->momx;
    momy = th->momy;
    momz = th->momz;

    // check if the rocket hits the wall near the bot
    dont_damage_player = true;
    hit_player = false;
    for (i=0; i<6; i++)
    {
      if (!P_TryMove (th, th->x + momx, th->y + momy, true) ) // a rocket hits
      {                                                 // something
        if (hit_player) goto hit; // a rocket hits a bot or the player
        delete th;
        return false;
      }
      th->momx = momx;
      th->momy = momy;
      th->momz = momz;
    }
hit:
    dont_damage_player = false;

    th->x = source->x;
    th->y = source->y;
    th->z = source->z + 4*8*FRACUNIT;

    P_CheckMissileSpawn (th);
    return true;
}


bool B_SpawnPlasma ( mobj_t*       source,
                     mobj_t*       dest )
{
    mobj_t*     th;
    angle_t     an;
    int         dist;

    th = P_SpawnMobj (source->x,
                      source->y,
                      source->z + 4*8*FRACUNIT, MT_PLASMA);
	
	if (th->info->seesound)
		S_Sound (th, CHAN_VOICE, th->info->seesound, 1, ATTN_NORM);

    th->target = source;        // where it came from

    dist = P_AproxDistance (dest->x - source->x, dest->y - source->y);
    dist = dist / th->info->speed;
    dist = P_AproxDistance (dest->x + dest->momx*dist - source->x,
                            dest->y + dest->momy*dist - source->y);
    dist = dist / th->info->speed;

    // pridict player position
    an = R_PointToAngle2 (source->x, source->y,
        dest->x + dest->momx*dist, dest->y + dest->momy*dist);

    if (dist < 1)
     dist = 1;

    th->angle = an;
    an >>= ANGLETOFINESHIFT;
    th->momx = FixedMul (th->info->speed, finecosine[an]);
    th->momy = FixedMul (th->info->speed, finesine[an]);
    th->momz = (dest->z - source->z) / dist;

    P_CheckMissileSpawn (th);
    return true;
}


//
// Changes bot weapon
//
void B_ChangeWeapon (mobj_t* mobj)
{
 int       i;
 int       change;
 Bot_t*    actor;

 actor = &bots[mobj->botnum];

 actor->SSGshot = actor->BFGshot = false; // Crashes without it. Why?

 for (change = 0; change < 256; change++)
 {
    i = P_Random(pr_botfire);

    if (i>=0 && i<5 && actor->sg && actor->shells > 0) actor->weapon = 2;
    if (i>=5 && i<70 && actor->ssg && actor->shells >= 2) actor->weapon = 3;
    if (i>=70 && i<100 && actor->chaingun && actor->bullets > 0) actor->weapon = 4;
    if (i>=100 && i<165 && actor->rl && actor->rockets > 0) actor->weapon = 5;
    if (i>=165 && i<=200 && actor->plasmagun && actor->cells > 0) actor->weapon = 6;
    if (i>=200 && i<256 && actor->bfg && actor->cells >= 40) actor->weapon = 7;
 }

  if (actor->ssg && actor->shells >= 4 && actor->weapon==2) actor->weapon = 3;

     if ((actor->shells <= 0 && actor->rockets <= 0
         && actor->cells <= 0 && actor->bullets >= 1 && !actor->chaingun)
         ||
         (actor->shells <= 0 && actor->rockets <= 0
         && actor->cells < 40 && actor->bullets >= 1 && !actor->plasmagun  && !actor->chaingun)
         ||
         (actor->shells < 2 && actor->rockets <= 0
         && actor->cells <= 0 && actor->bullets >= 1 && !actor->sg  && !actor->chaingun))
       {
         actor->weapon = 1;
         return;
       }

     if ((actor->shells <= 0 && actor->rockets <= 0
       && actor->cells <= 0 && actor->bullets <= 0)
       ||
        (actor->shells <= 0 && actor->rockets <= 0
        && actor->cells < 40 && actor->bullets <= 0 && !actor->plasmagun)
       ||
        (actor->shells <= 2 && actor->rockets <= 0
        && actor->cells <= 0  && actor->bullets <= 0 && !actor->sg))
       {
         if (actor->chainsaw) actor->weapon = 0;
		 else actor->weapon =-1;
       }
}

void BOT_Attack (mobj_t* actor)
{
 int         damage;
 int         slope;
 int         angle;
 Bot_t*      b;
 int         i;

b = &bots[actor->botnum];

B_FaceTarget (actor);

angle = actor->angle;
slope = P_AimLineAttack (actor, angle, 16*64*FRACUNIT);


//---------------------ChainSaw----------------------------------------

if (b->weapon==0)
{
   if (P_AproxDistance (actor->target->x - actor->x,
                        actor->target->y - actor->y)
			> 180*FRACUNIT) return;
   if (b->ChainsawCounter==0)
   {
     int			t;

	 damage = 2 * (P_Random (pr_botfire)%10+1);
	 t = P_Random (pr_botfire);
	 angle += (t - P_Random (pr_botfire)) << 18;
	
	 // use meleerange + 1 so the puff doesn't skip the flash
	 P_LineAttack (actor, angle, MELEERANGE+1,
				   P_AimLineAttack (actor, angle, MELEERANGE+1), damage);

	 if (!linetarget)
	 {
		 S_Sound (actor, CHAN_WEAPON, "weapons/sawfull", 1, ATTN_NORM);
		 return;
	 }
	 S_Sound (actor, CHAN_WEAPON, "weapons/sawhit", 1, ATTN_NORM);
	 b->ChainsawCounter = 5;
   }
   b->ChainsawCounter--;
}

//---------------------Pistol-----------------------------------------
if (b->weapon == 1)
  {
    if (b->bullets > 0)
      {
        if (b->PistolCounter==0)
          {
	        S_Sound (actor, CHAN_WEAPON, "weapons/pistol", 1, ATTN_NORM);

            damage = 5*(P_Random (pr_botfire)%3+1);
            angle += 2*((P_Random(pr_botfire)-P_Random(pr_botfire))<<18);
            P_LineAttack (actor, angle, MISSILERANGE, slope, damage);

            b->PistolCounter=14;
            b->bullets--;

            P_SetMobjState (actor, S_PLAY_ATK2);
          }

        b->PistolCounter--;
      }
    else
      B_ChangeWeapon (actor);
  }

//---------------------ShotGun----------------------------------------
if (b->weapon == 2)
  {
    if (b->shells > 0)
      {
        if (b->SGCounter==0)
          {
         	S_Sound (actor, CHAN_WEAPON,  "weapons/shotgf", 1, ATTN_NORM);

            for (i=0 ; i<5 ; i++)
               P_GunShot (actor, false);

            b->SGCounter=36;
            b->shells--;
            P_SetMobjState (actor, S_PLAY_ATK2);
          }

        b->SGCounter--;
      }
    else
      B_ChangeWeapon (actor);
  }

//---------------------SuperShotGun-----------------------------------

if (b->weapon == 3  || b->SSGshot)
{
    if (b->shells >= 2 || b->SSGshot)
    {
        if (b->SSGCounter==43)
        {
            S_Sound (actor, CHAN_WEAPON, "weapons/sshotf", 1, ATTN_NORM);

            if (actor->target->momx || actor->target->momy)
              angle += 1.4*((P_Random(pr_botfire)-P_Random(pr_botfire))<<19);

            for (i=0 ; i<20 ; i++)
            {
              damage = 5*(P_Random(pr_botfire)%3+1);
        	  angle += (P_Random (pr_botfire)-P_Random (pr_botfire)) << 19;
              P_LineAttack (actor, angle, MISSILERANGE,
                            slope + ((P_Random(pr_botfire)-P_Random(pr_botfire))<<5), damage);
            }

            b->shells -= 2;
            b->SSGshot = true;

            P_SetMobjState (actor, S_PLAY_ATK2);
        }

        // reload ssg
		if (b->SSGCounter==27)
          S_Sound (actor, CHAN_WEAPON, "weapons/sshoto", 1, ATTN_NORM);        
        if (b->SSGCounter==14)
	      S_Sound (actor, CHAN_WEAPON, "weapons/sshotl", 1, ATTN_NORM);      
        if (b->SSGCounter==2)
	      S_Sound (actor, CHAN_WEAPON, "weapons/sshotc", 1, ATTN_NORM);

        b->SSGCounter--;

        if (b->SSGCounter<=0)
		{
           b->SSGCounter=51;
		   b->SSGshot = false;
		}

    }
    else
      B_ChangeWeapon (actor);
}

//---------------------ChainGun---------------------------------------

if (b->weapon == 4)
  {
    if (b->bullets > 0)
      {
        if (b->ChainGunCounter==0)
          {
           	  S_Sound (actor, CHAN_WEAPON, "weapons/chngun", 1, ATTN_NORM);

              damage = 5*(P_Random (pr_botfire)%3+1);
              angle += 3*((P_Random(pr_botfire)-P_Random(pr_botfire))<<18);
              P_LineAttack (actor, angle, MISSILERANGE, slope, damage);

              b->ChainGunCounter=4;
              b->bullets--;

              P_SetMobjState (actor, S_PLAY_ATK2);
          }

        b->ChainGunCounter--;
      }
    else
      B_ChangeWeapon (actor);
  }

//---------------------Rocket Launcher--------------------------------

 if (b->weapon == 5)
  {
    if (b->rockets > 0)
      {
        if (b->RLCounter==0)
          {
            if (!B_SpawnRocket (actor, actor->target, MT_ROCKET))
            {
              // impossible to shoot. The rocket will hit a wall near
              // the bot!
              b->RLCounter=4;
              return;
            }

            b->RLCounter=25;
            b->rockets--;

            P_SetMobjState (actor, S_PLAY_ATK2);
          }
        b->RLCounter--;
      }
    else
      B_ChangeWeapon (actor);
  }

// ---------------Plasmagun--------------------------------------------

 if (b->weapon == 6)
  {
    if (b->cells > 0)
      {
        if (b->PlasmagunCounter==0)
          {
            B_SpawnPlasma (actor, actor->target);
            b->PlasmagunCounter=4;
            b->cells--;
            P_SetMobjState (actor, S_PLAY_ATK2);
          }

        b->PlasmagunCounter--;
      }
    else
      B_ChangeWeapon (actor);
  }

//---------------------BFG9000-----------------------------------
  if (b->weapon == 7 || b->BFGshot)
  {
	  if (b->cells >= 40)
         B_FireBFG (actor);
	  else
		 B_ChangeWeapon (actor);
  }
}

void B_FireBFG (mobj_t* actor)
{
  mobj_t*     th;
  angle_t     an;
  int         i,j;
  bool        wall=false;
  fixed_t     bestdist=1000*FRACUNIT;
  fixed_t     dist;
  angle_t     bestang=0;
  Bot_t*      b;

  b = &bots[actor->botnum];


  if (b->BFGCounter==42) 
	  S_Sound (actor, CHAN_WEAPON, "weapons/bfgf", 1, ATTN_NORM);
  b->BFGCounter--;
  b->BFGshot = true;

  if (b->BFGCounter==0)
  {
	th = P_SpawnMobj (actor->x,
                      actor->y,
                      actor->z + 4*8*FRACUNIT, MT_BFG);

	if (th->info->seesound)
		S_Sound (th, CHAN_VOICE, th->info->seesound, 1, ATTN_NORM);

    th->target = actor;        // where it came from
    an = actor->angle;
    
	an -= ANG90/2 - ANG90/9;
    
	dont_damage_player = true;
    // find the nearest wall to shoot at
    for (i=0; i<7; i++) // try 7 different angles
    {
      th->momx = FixedMul (th->info->speed, finecosine[an>>ANGLETOFINESHIFT]);
      th->momy = FixedMul (th->info->speed, finesine[an>>ANGLETOFINESHIFT]);

	  for (j=0; j<6; j++)
        if (!P_TryMove (th, th->x+th->momx, th->y+th->momy, true))
        {
           wall = true;
           dist = P_AproxDistance (actor->x - th->x, actor->y - th->y);
           if (bestdist > dist)
           {
             bestdist = dist;
             bestang = an;
			 goto nextdir;
           }
        }

  nextdir: 
	   an += ANG90/9;
	   th->UnlinkFromWorld();
       th->x = actor->x;
       th->y = actor->y;
	   th->LinkToWorld();
    }
    
	dont_damage_player = false;


    if (wall) an = th->angle = bestang;
    else  an = th->angle = actor->angle;

	th->momx = FixedMul (th->info->speed, finecosine[an>>ANGLETOFINESHIFT]);
    th->momy = FixedMul (th->info->speed, finesine[an>>ANGLETOFINESHIFT]);
	th->momz = 0;

    P_CheckMissileSpawn (th);


    b->BFGshot = false;
    b->BFGCounter=45;
    b->cells -= 40;
    P_SetMobjState (actor, S_PLAY_ATK2);
  }


}

void B_ChainsawReady(mobj_t* actor)
{	  
   Bot_t*      b;
   b = &bots[actor->botnum];

   if (b->weapon==0 && (!actor->target 
	   || P_AproxDistance (actor->target->x - actor->x,
                           actor->target->y - actor->y)
		    > 180*FRACUNIT) )
   {
      if (!b->ChainsawCounter)
	  {  
		S_Sound (actor, CHAN_WEAPON, "weapons/sawidle", 1, ATTN_NORM);
		b->ChainsawCounter=6;
	  }
	  b->ChainsawCounter--;
   }
}