// Bot include
#include "bot.h"
#include "bot_look.h"
#include "bot_move.h"
#include "bot_fire.h"
#include "bot_node.h"
#include "bot_ctf.h"
// Doom include
#include "r_state.h"
#include "v_video.h"
#include "m_argv.h"


void B_LookForTarget(mobj_t* mobj);

Bot_t     bots[MAXBOTS];
int       botnumber=0;

#define RespawnTime 80
int     RespawnCounter = RespawnTime;

bool ctf;


char* botnames[MAXBOTS] = {
 "Frag-God",
 "Thresh",
 "Reptile",
 "Archer",
 "Freak",
 "TF-Master",
 "John Carmack",
 "Quaker",
 "FragMaster",
 "Punisher",
 "Romero",
 "Xoleras",
 "Bot1",
 "Bot2",
 "Bot3",
 "Bot4"
};

char colors[11][9] = 
{
   "40 cf 00",    //0  = Green
   "b0 b0 b0",    //1  = Gray
   "bf 8f 2f",    //2  = Indigo
   "8f 00 00",    //3  = Deep Red
   "ff ff ff",    //4  = White
   "ff bf 4f",    //5  = Bright Brown
   "bf 00 00",    //6  = Red
   "00 00 ff",    //7  = Blue
   "00 00 7f",    //8  = Dark Blue
   "ff ff 00",    //9  = Yellow
   "cf df 90"     //10 = Bleached Bone
};

void InitBots(void)
{  
    multiplayer = true;

	if (Args.CheckParm("-ctf") )
	{
        redflagreturn = 0;
        blueflagreturn = 0;		
		ctf = true;
	}
	else
		ctf = false;

	for (int i=0, p=9; i<11; i++, p++)
	{
    	R_BuildPlayerTranslation (p, V_GetColorFromString (NULL, colors[i]));
	}
    
	for (i=0; i<MAXBOTS; i++)
	{
	  if (i<10) bots[i].color = i+1;
	  else bots[i].color = i-10;
	}

    if (ctf)
	  for (int i=0; i<MAXBOTS; i++) 
  	    bots[i].color = 6;
}

//
// add a bot
//
BEGIN_COMMAND (addbot)
{
	if (consoleplayer != Net_Arbitrator)
    {
		Printf (PRINT_HIGH, "Only player %d can add bots\n", Net_Arbitrator + 1);
        return;
    }
    
	Net_WriteByte (DEM_ADDBOT);
}
END_COMMAND (addbot)

void B_AddBot(void)
{
    if (botnumber>=MAXBOTS) 
	{
	    Printf (PRINT_HIGH, "You can add only 16 bots\n");
		return; 
	}

	if (ctf && !strcmp(bots[botnumber].team, "blue") )
		BOT_SpawnBlueBot(botnumber);
	else
        BOT_DeathmatchSpawnBot (botnumber);
	
	bots[botnumber].frags = 0;
    Printf (PRINT_HIGH, "%s has entered the game\n", botnames[botnumber]);

    botnumber++;
}


//
// remove all bots
//
BEGIN_COMMAND (removebots)
{
	if (consoleplayer != Net_Arbitrator)
    {
		Printf (PRINT_HIGH, "Only player %d can remove bots\n", Net_Arbitrator + 1);
        return;
    }
    Net_WriteByte (DEM_KILLBOTS);
}
END_COMMAND (removebots)

void RemoveAllBots (void)
{
	for (int i=0; i<botnumber; i++)
      delete bots[i].mo;
    botnumber = 0;
    Printf (PRINT_HIGH, "Removed all bots\n");
}


BEGIN_CUSTOM_CVAR (botteam, "red", CVAR_SERVERINFO)
{  
  if (ctf)
  {
    if (stricmp(var.string, "red") && stricmp(var.string, "blue") )
	{
		Printf(PRINT_HIGH, "Choose red or blue team!\n");
		return;
	}
  }
  
  for (int i=botnumber; i<MAXBOTS; i++)
    sprintf (bots[i].team, var.string);
}
END_CUSTOM_CVAR (botteam)


BEGIN_CUSTOM_CVAR (botcolor, "0", CVAR_SERVERINFO)
{
  if (ctf)
  {
	Printf(PRINT_HIGH, "You can't change bot color in ctf\n");
	return;
  }
		
  int c = (int)var.value;

  if (c>10) c=10;
  if (c<0) c=0;

  for (int i=botnumber; i<MAXBOTS; i++) 
	 bots[i].color = c;
}
END_CUSTOM_CVAR (botcolor)


// ctf commands
BEGIN_COMMAND (botguard)
{
  for (int i=0; i<botnumber; i++)
	if (   P_CheckSight(players[consoleplayer].mo, bots[i].mo) 
		&& bots[i].mo->IsTeammate (players[consoleplayer].mo))
	{
		bots[i].guard = true;
		B_ClearBotRoute (bots[i].mo);
	}
}
END_COMMAND (botguard)

BEGIN_COMMAND (botfree)
{
  for (int i=0; i<botnumber; i++)
	if (   P_CheckSight(players[consoleplayer].mo, bots[i].mo) 
		&& (bots[i].guard || bots[i].follow)
		&& bots[i].mo->IsTeammate (players[consoleplayer].mo))
	{
        bots[i].guard = false;
		bots[i].follow = false;
	}
}
END_COMMAND (botfree)

//
// MAIN BOT AI
//
void BOT_Thinker (mobj_t* mobj)
{
  Bot_t*    b;

  b = &bots[mobj->botnum];

  if (mobj->health > 0)
  {
/*---------------------The main part of the BOT AI--------------------------*/

      if (P_Random(pr_botmain) > 100 && !b->GoToThing)
      {
		 if (B_LookForThings (mobj)  )
         {
           b->GoToThing = true;
           B_ClearBotRoute(mobj);
         }
         else
         {
           b->GoToThing = false;
		   if ( altnode && !b->destnode 
			  && (!mobj->target || (mobj->target && b->weapon <= 1)) )
		   {
             b->destnode = B_FindThingNode( B_NeedThing(mobj) );
			 if (b->destnode)
			 {
			   B_FindRoutePos(mobj, b->destnode);
			   b->lastvisitednode = 0;
			 }
		   }
         }
      }

	  if (b->GoToThing)
	  {
		  if (!B_MoveToThing (mobj))
			  b->GoToThing = false;
	  }
	  
	  if (b->destnode && P_Random(pr_botmain) < 50)
		  B_FindRoutePos(mobj, b->destnode);

	  if (!b->GoToThing && !b->ontheway) mobj->mangle = mobj->angle;

	  B_LookForTarget(mobj);

	  if (b->guard)
	  {
		  b->forwardmove = 0;
		  goto fight_move;
	  }

      // nodes...
	  if (altnode &&
		  (   !mobj->target 
		   || (mobj->target && (b->weapon==-1 || b->weapon==1 || mobj->health <= 30))
		   || ctf
		  ) 
		 )
      {
          if (!b->ontheway && !b->GoToThing && node_count && P_Random(pr_botmain) < 30)
          {
			b->routeposition = B_FindAnyVisNode (mobj);
            if (b->routeposition) b->ontheway = true;
          }
      }

      // target is found! don't move to the next node!
      if (mobj->target && (b->weapon > 1 || b->weapon==0) && mobj->health > 30) // doesn't have enough ammo/health to fight
	  {
         if (!ctf)
		     B_ClearBotRoute(mobj);
	  }

	  if ((!mobj->target || b->GoToThing) && !b->ontheway)
        b->forwardmove = 50;

      // nodes...
	  if (altnode && b->ontheway)
		  B_MoveToNextAltNode (mobj);

fight_move:

	  // Bot has found the player or another bot!
      if (mobj->target || b->SSGshot || b->BFGshot)
         BOT_Attack (mobj);

      if (mobj->target && !b->GoToThing && !b->ontheway)
		  B_MoveFight (mobj);
      else
	   if (!B_Move (mobj) )  { B_NewChaseDir(mobj); b->GoToThing=false; }

	  // power ups
      if (b->powers[pw_invulnerability])
          b->powers[pw_invulnerability]--;

      if (b->powers[pw_invisibility])
          if (! --b->powers[pw_invisibility] )
              b->mo->flags &= ~MF_SHADOW;

	  B_CheckPickUp (mobj);

	  if (!mobj->target)
		  b->wait = 0;

	  if (!b->wait)
		  b->forwardmove = 0;
  
	  B_ChainsawReady(mobj);
  
/*--------------------------------------------------------------------------*/
  }

  // BOT is  D E A D!
  if (mobj->health <= 0)
  {
	  if (ctf && b->flag)
	  {
        AActor*  mo;

		if (!stricmp(b->team, "red") )
		{
			mo = new AActor(mobj->x, mobj->y, 
				            mobj->z, MT_BFLAG);
		}		
		if (!stricmp(b->team, "blue") )
		{
			mo = new AActor(mobj->x, mobj->y, 
				            mobj->z, MT_RFLAG);
		}
		
		mo->momz += 10*FRACUNIT;
		b->flag = false;
	  }
	
	  RespawnCounter--;
      if (RespawnCounter==0)
      {
        RespawnCounter = RespawnTime;
        mobj->alive = false;
        
		if (ctf && !strcmp(b->team, "blue") )
			BOT_SpawnBlueBot(mobj->botnum);
		else
		    BOT_DeathmatchSpawnBot (mobj->botnum);
      }
  }  


  if (Args.CheckParm("-testnodes"))
  {
    if (!node_count) 
		return; // no nodes
	if ( (P_Random(pr_botmain))%10 != 0)
		return;

    for (int i=1; i<=node_count; i++)
	{
      if (abs(players[0].mo->x - botnodes[i].mo->x) <= 20*FRACUNIT &&
          abs(players[0].mo->y - botnodes[i].mo->y) <= 20*FRACUNIT)
	  {
         Printf(PRINT_HIGH, "You are near node %d\n", i);
	  }
	}
	
	Printf(PRINT_HIGH, "next = %d  ", b->routeposition);	
	Printf(PRINT_HIGH, "dest = %d\n", b->destnode);
  }




} // end of BOT_Thinker


void B_LookForTarget(mobj_t* mobj)
{
   // looking for targets...
   if (   !mobj->target || mobj->target->health<=0 // no target or it's dead
       || !P_CheckSight (mobj, mobj->target) ) // doesn't see the target
   { // then look for the target...
	  if (P_Random(pr_botmain) > 100)
	  {
        if (!P_LookForPlayers(mobj, false) ) B_LookForBots (mobj);
	  }
      else
        if (!B_LookForBots (mobj) ) P_LookForPlayers(mobj, false);
   }
   if (Args.CheckParm("-testnodes")) mobj->target = 0;
}


//
// Check if Bot spawn start is free (doesn't work correctly)
//
bool BOT_CheckSpot (mapthing2_t*   mthing)
{
    fixed_t             x,y;
    mobj_t*             mo;
    subsector_t*        ss;
    unsigned            an;

    x = mthing->x << FRACBITS;
    y = mthing->y << FRACBITS;

    // spawn a teleport fog
    ss = R_PointInSubsector (x,y);
    an = ( ANG45 * (mthing->angle/45) ) >> ANGLETOFINESHIFT;

    mo = P_SpawnMobj(x+20*finecosine[an], y+20*finesine[an]
                     , ss->sector->floorheight
                     , MT_TFOG);

    if (level.time)
      S_Sound (mo, CHAN_VOICE, "misc/teleport", 1, ATTN_NORM);   // don't start sound on the first frame

    return true;
}


//
// BOT_SpawnBot
//
void BOT_SpawnBot (mapthing2_t* mthing, int botnum)
{
    Bot_t *b = &bots[botnum];

    fixed_t x = mthing->x << FRACBITS;
    fixed_t y = mthing->y << FRACBITS;

    b->mo = P_SpawnMobj (x,y,ONFLOORZ, MT_DMBOT);

    // telefrag anything in this spot
    P_TeleportMove (b->mo, b->mo->x, b->mo->y, b->mo->z, true);

    b->mo->botnum = botnum;

    if (ctf)
	{
		if (!strcmp(b->team, "blue") )
			b->color = 7;
		else 
			b->color = 6;
	}

	int color = b->color+9;
	b->mo->translation = translationtables + 256*color;

    b->mo->angle = ANG45 * (mthing->angle/45);
    b->lastvisitednode = 0;
    b->armor=0;
    b->avoidingrocket = false;
	b->wait = 0;
	b->follow = false;
	b->guard = false;

	b->mo->alive = true;
    
	b->bullets = 50;
    b->shells = 0;
	b->cells = 0;
	b->rockets = 0;

    b->weapon = 1;  // Pistol

	b->sg = false;
	b->ssg = false;
	b->rl = false;
	b->chaingun = false;
	b->plasmagun = false;
	b->bfg = false;

    b->ChainsawCounter = 5;
    b->PistolCounter = 14;
    b->SGCounter = 36;
    b->SSGCounter = 51;
    b->PlasmagunCounter = 4;
    b->ChainGunCounter = 4;
    b->RLCounter = 25;
    b->BFGCounter = 45;

	b->SSGshot = false; 
	b->BFGshot = false;

	b->flag = false;
	
	B_ClearBotRoute(b->mo);
}


//
// BOT_DeathmatchSpawnBot
//
void BOT_DeathmatchSpawnBot (int botnum)
{
    int             i,j;
    int             selections;

    selections = deathmatch_p - deathmatchstarts;

    for (j=0 ; j<32 ; j++)
    {
        i = P_Random(pr_botspawnmobj) % selections;
        if (BOT_CheckSpot (&deathmatchstarts[i]) )
        {
            BOT_SpawnBot (&deathmatchstarts[i], botnum);
            return;
        }
    }
}

//
// BOT_SpawnBlueBot
// For CTF. Spawn a bot on a random player start.
//
void BOT_SpawnBlueBot (int botnum)
{
    int             i,j;

    for (j=0 ; j<20 ; j++)
    {
        i = P_Random(pr_botspawnmobj) % 3;
        if (BOT_CheckSpot (&playerstarts[i]) )
        {
            BOT_SpawnBot (&playerstarts[i], botnum);
            return;
        }
    }
}