// Bot_move.c
// Movements of the bot

#include "bot.h"
#include "am_map.h"
#include "bot_move.h"
#include "bot_node.h"
#include "p_lnspec.h"

#include <math.h>

extern  line_t **spechit;
extern  int      numspechit;


//
// P_Thrust
// Moves the given origin along a given angle.
//
void P_Thrust ( mobj_t*       mo,
                angle_t       angle,
                fixed_t       move )
{
    angle >>= ANGLETOFINESHIFT;

    mo->momx += FixedMul(move,finecosine[angle]);
    mo->momy += FixedMul(move,finesine[angle]);
}


//
// Move a bot
//
bool B_Move (mobj_t* actor)
{
  angle_t       angle;
  bool          onground;
  Bot_t*        b;
  
  b = &bots[actor->botnum];

  if (!actor->momx && !actor->momy && !actor->target) 
    P_SetMobjState (actor, S_PLAY);

  // Reactiontime is used to prevent movement
  // for a bit after a teleport.
  if (actor->reactiontime)
  {
    actor->reactiontime--;
    return true;
  }

  if (!b->forwardmove && !b->sidemove) 
	  return true;

  if (!actor->target || b->GoToThing || b->ontheway) 
	  b->sidemove=0;

  angle = actor->mangle;
  P_Thrust (actor, angle, b->forwardmove*2048);

  angle = actor->mangle - ANG90;
  P_Thrust (actor, angle, b->sidemove*2048);


// Check move
  if (!P_CheckPosition (actor, actor->x+actor->momx, actor->y+actor->momy))
    return false;           // solid wall or thing

  if (tmceilingz - tmfloorz < actor->height)
  {
     if (!numspechit)  return false;
     while (numspechit--)
     {
        line_t* ld = spechit[numspechit];
        if (P_ActivateLine (ld, actor, 0, SPAC_USE) ||
			P_ActivateLine (ld, actor, 0, SPAC_PUSH) )
          return true;
     }

     return false;       // doesn't fit and not a door
  }

  if ( !(actor->flags&MF_TELEPORT)
       &&tmceilingz - actor->z < actor->height)
  {
     return false;       // mobj must lower itself to fit
  }

  if ( !(actor->flags&MF_TELEPORT)
       && tmfloorz - actor->z > 24*FRACUNIT ) // too big a step up
  {
      if (tmfloorz - actor->z > 45*FRACUNIT)
          return false;   // too big to jump
      else
      {
         onground = (actor->z <= actor->floorz);
         if (onground)
         {
     		actor->momz += 8*FRACUNIT;
//			S_Sound (actor, CHAN_BODY, "*jump1", 1, ATTN_NORM);
         }
      }
  }


  // change state
  if (actor->state == &states[S_PLAY])
    P_SetMobjState (actor, S_PLAY_RUN1);

  // move is ok
  return true;
}


mobj_t* B_SeeMissile (mobj_t* actor)
{
 Bot_t*      b;
 mobj_t*     missile=0;
 int         i=0;
 fixed_t     dist=0;
 fixed_t     x,y,m_momx, m_momy, mx, my, momx, momy;
 mobj_t*     mo;
 
 b = &bots[actor->botnum];

 x = actor->x;
 y = actor->y;

 // scan for the closest missile
 TThinkerIterator<AActor> iterator;
 while ( (mo = iterator.Next() ) )
 {
   fixed_t  d;

   if ( !(mo->flags & MF_MISSILE) )
         continue;  // it is not a missile..
   if (mo->target == actor) // it is actor's missile
	   continue;

   // see if the missile is heading our way
   d = P_AproxDistance (x - mo->x, y - mo->y);
   if ( d > P_AproxDistance (x - mo->x+mo->momx, y - mo->y+mo->momy))
      continue;

   if (i==0 || dist > d)
   {
     missile = mo;
     dist = d;
   }

   i++;
 }

   if (!missile) 
   {	   
	   b->avoidingrocket = false;
	   return 0;
   }

   momx = actor->momx;
   momy = actor->momy;

   m_momx = missile->momx;
   m_momy = missile->momy;
   mx = missile->x;
   my = missile->y;

   for (i=1; i<300; i++)
   {
     if ( P_AproxDistance (   (x+momx*i) - (mx+m_momx*i),
                              (y+momy*i) - (my+m_momy*i) )
           <= (5*missile->radius + actor->radius) )
       return missile;
   }

   b->avoidingrocket = false;
   return 0;

}

void B_AvoidMissile (mobj_t* actor, mobj_t* missile)
{
   if (!missile) return;
   
   Bot_t* b = &bots[actor->botnum];

   if (!b->avoidingrocket)
   {
     actor->mangle = missile->angle;
     if (b->sidemove >= 0) b->sidemove = -50;
     else b->sidemove = 50;
     b->avoidingrocket = true;
   }

   if (!B_Move(actor))
   {
     if (b->sidemove > 0  )  b->sidemove = -50;
     else                    b->sidemove = 50;
   }
}

//
// B_HellSlime
//
bool B_HellSlime(fixed_t x, fixed_t y)
{
  subsector_t *subsec;
  subsec = R_PointInSubsector (x, y);
  int special = subsec->sector->special & ~SECRET_MASK;

  if (special >= dLight_Flicker && special <= dLight_FireFlicker)
  {
	switch (special)
	{
	  case dDamage_Hellslime:
		  return true;
	  case dDamage_Nukage:
		  return true;
	  case dDamage_SuperHellslime:
		  return true;
	  case dLight_Strobe_Hurt:
		  return true;
	}
  }

  return false;
}

//
// B_MoveFight
// Movements of the bot during fight
//
void B_MoveFight (mobj_t* actor)
{
 Bot_t*      b;
 fixed_t     dist=0;
 int         p;
 int         frand, nrand;
 
 b = &bots[actor->botnum];

  B_AvoidMissile (actor, B_SeeMissile (actor));
  if (b->avoidingrocket)
	  return;

  fixed_t SHOOTDIST_FAR = 300*FRACUNIT;
  fixed_t SHOOTDIST_NEAR = 50*FRACUNIT;
  frand = 50; nrand = 100;

  if (b->wait)
	  b->wait--;
  else
  {
    // ssg
    if (b->weapon == 3)  SHOOTDIST_FAR = 150*FRACUNIT;
    if (b->weapon <= 0) 
	{
      SHOOTDIST_FAR = 0;
      SHOOTDIST_NEAR = 0;
      frand = 40;
	}

    dist = P_AproxDistance (actor->target->x - actor->x,
                            actor->target->y - actor->y);
 
    if (dist > SHOOTDIST_FAR && P_Random(pr_botmove) > frand) b->forwardmove = 50;
    if (dist < SHOOTDIST_NEAR && P_Random(pr_botmove) < nrand) b->forwardmove = -50;
  }

  if (b->sidemove == 0)
  {
    if (P_Random(pr_botmove) < 128)  b->sidemove = 50;
    else                             b->sidemove = -50;
  }

  //
  // avoid hell slime
  //
  angle_t  angle;
  fixed_t  x, y;
  fixed_t  dx, dy;
  bool     lava = false;

  angle = actor->mangle;
  angle >>= ANGLETOFINESHIFT;
  dx = FixedMul(2048*b->forwardmove, finecosine[angle]);
  dy = FixedMul(2048*b->forwardmove, finesine[angle]);
  angle = actor->mangle - ANG90;
  angle >>= ANGLETOFINESHIFT;
  dx += FixedMul(2048*b->sidemove, finecosine[angle]);
  dy += FixedMul(2048*b->sidemove, finesine[angle]);

  x = actor->x;
  y = actor->y;

  if (!B_HellSlime(x, y) )
  {
     if (   B_HellSlime(x + 40*dx, y + 40*dy)
		 && B_HellSlime(x + 120*dx, y + 120*dy)
		 && B_HellSlime(x + 200*dx, y + 200*dy) )
		 lava = true;
  }

  if (lava)
  {
	  if (b->forwardmove != 0 && !b->ontheway)
	  {
	    if (b->forwardmove > 0)  b->forwardmove = -50;
        else                     b->forwardmove = 50;

		b->wait = 30;
	  }

	  if (b->sidemove > 0  )  b->sidemove = -50;
      else                    b->sidemove = 50;
  }
  // end of the lava check


  p = P_Random(pr_botmove);

  if (!B_Move(actor) || p%10==0)
  {
    if (b->sidemove > 0  )  b->sidemove = -50;
    else                    b->sidemove = 50;
  }

  p = P_Random(pr_botmove);
  if (b->weapon<=0 && p<220) b->sidemove = 0;

}


//
// B_FaceTarget
//
void B_FaceTarget (mobj_t* actor)
{
 if (!actor->target) return;
 actor->angle = R_PointToAngle2 (actor->x,
                                 actor->y,
                                 actor->target->x,
                                 actor->target->y);
}


// Change direction of a bot
void B_NewChaseDir (mobj_t* actor)
{
   actor->mangle = actor->angle = 8*P_Random(pr_botnewchasedir)<<21;
}

//
// Check bot's movement from point to point
//
bool
B_CheckMovePtoP (mobj_t* actor, mobj_t* dest, mobj_t* target)
{
 fixed_t  x,y,xt,yt,a,b, oldfloorz;
 int      checkpoints;
 int      i;

 x = dest->x;
 y = dest->y;
 xt = target->x;
 yt = target->y;

 checkpoints = FixedDiv(P_AproxDistance (xt-x, yt-y), 25*FRACUNIT);
 checkpoints >>= 16;

 if (checkpoints) 
 {
   a = (xt - x) / checkpoints;
   b = (yt - y) / checkpoints;
 }
 else
 {
   a = xt - x;
   b = yt - y;
   checkpoints = 1;
 }

 // gets a floor z at (x,y)
 P_CheckPosition (actor, x, y);
 oldfloorz = tmfloorz;

 for (i=0; i < checkpoints; i++)
 {
   x += a;
   y += b;

   if (!P_CheckPosition (actor, x, y))
	   return false;  // solid wall

   if ((tmfloorz - oldfloorz) > 45*FRACUNIT )
       return false;  // too big to jump over

   if (tmceilingz - tmfloorz < actor->height && tmceilingz-tmfloorz != 0 && !numspechit)
     return false;

   oldfloorz = tmfloorz;
 }

 return true;
}

//
// Move a bot to a thing
//
bool B_MoveToThing (mobj_t* actor)
{
   Bot_t*  b = &bots[actor->botnum];
   
   if (!b->thing)
	   return false;

   if (!P_CheckSight (actor, b->thing) )
       return false;

   if (!B_CheckMovePtoP (actor, actor, b->thing))
       return false;

   actor->mangle = R_PointToAngle2 (actor->x, actor->y,
                                    b->thing->x, b->thing->y);
   if (!actor->target) 
	   actor->angle = actor->mangle;
   
   // move is ok
   return true;
}

