#include "bot.h"
#include "bot_node.h"
#include "bot_move.h"
#include "bot_look.h"

#include "m_argv.h"
#include "m_misc.h"
#include "i_system.h"
#include "z_zone.h"

botnode_t botnodes[MAXBOTNODES];
// number of nodes on the map
int   node_count;

void B_ClearBotRoute(mobj_t* actor)
{
  Bot_t*    b;

  b = &bots[actor->botnum];

  b->ontheway = false;
  b->routeposition = 0;
  b->destnode = 0;
  b->roam = -1;
}


/////////////////////////////////////////////////////////////////////////////
/////                       NEW NODE SYSTEM                             /////
/////////////////////////////////////////////////////////////////////////////
bool    altnode;
byte*   nodebase;


void B_InitAltNodes (void)
{
 
  if (Args.CheckParm ("-nodes"))
  {
     altnode = true;
     Printf (PRINT_HIGH, "Init node system...\n");     

	 DArgs *files = Args.GatherFiles ("-nodes", ".nod", false);
     if (files->NumArgs() > 0)
     {   
		 LoadNodebase(files->GetArg (0));
		 delete files;
		 return;
	 }
	 else
       I_Error ("No nodebase file...");        
  }
  else 
	 altnode = false;
}


void LoadNodebase (char *filename)
{
    fixed_t*            get;
    int                 i;
    byte*               savebuffer;
    fixed_t             x, y, angle;
    botnode_t*          node;

	node_count = 0;

    if (Args.CheckParm("-testnodes") || Args.CheckParm("-routebuilder") )
    {  
	  // nodes will be visible
	  mobjinfo[MT_BOTNODE].flags &= ~MF_NOBLOCKMAP;
      mobjinfo[MT_BOTNODE].flags2 &= ~MF2_DONTDRAW;
	}

    M_ReadFile (filename, &savebuffer);
    nodebase = savebuffer;

    get = (fixed_t *)nodebase;

    node_count = *get++;
	Printf (PRINT_HIGH, "%d nodes on the map\n", node_count);

    for (i=1; i<=node_count; i++)
    {
      // load data...
      x = *get++;
      y = *get++;
      angle = *get++;
      botnodes[i].special1 = *get++;
      botnodes[i].special2 = *get++;
      // and spawn a node...
      node = &botnodes[i];
      node->mo = P_SpawnMobj (x, y, ONFLOORZ, MT_BOTNODE);
      node->mo->botnum = i;
      node->mo->angle = angle;
    }


    Z_Free(savebuffer);
}

//
// B_FindNextVisNodes
//
int B_FindNextVisNode (mobj_t* actor)
{
 Bot_t*    b;

 b = &bots[actor->botnum];

 if (b->lastvisitednode)
 {
   if (   b->lastvisitednode+1 <= node_count
       && B_CheckMovePtoP (actor, actor, botnodes[b->lastvisitednode+1].mo))
     return b->lastvisitednode+1;
   
   if (   b->lastvisitednode+2 <= node_count
       && B_CheckMovePtoP (actor, actor, botnodes[b->lastvisitednode+2].mo))
     return b->lastvisitednode+2;
   
   if (   b->lastvisitednode+3 <= node_count
       && B_CheckMovePtoP (actor, actor, botnodes[b->lastvisitednode+3].mo))
     return b->lastvisitednode+3;         
 }

 return 0;
}

//
// B_FindAnyVisNode
//
int B_FindAnyVisNode (mobj_t* actor)
{
 int       node;
 Bot_t*    b;

 b = &bots[actor->botnum];

 if (b->lastvisitednode && b->lastvisitednode+1 <= node_count)
 {
   node = b->lastvisitednode + 1;	   
   if (   botnodes[node].special2 != 1 
	   && P_CheckSight (actor, botnodes[node].mo) 
	   && B_CheckMovePtoP (actor, actor, botnodes[node].mo) )       
     return node;
 }

 for (node=node_count; node>=1; node--)
 {
   if (node == b->lastvisitednode + 1)
	   continue;
   if (!P_CheckSight (actor, botnodes[node].mo) )
       continue;
   if (botnodes[node].special2 == 1 && botnodes[node].special1 != 5)
	   continue;
   if (!B_CheckMovePtoP (actor, actor, botnodes[node].mo))
       continue;           // impossible to get to this node
   return node;
 }

 return 0;
}

//
// B_FindAnyVisNode2
// 
int B_FindAnyVisNode2 (mobj_t* actor)
{
 int       node;
 Bot_t*    b;

 b = &bots[actor->botnum];

 // !!!!!!!!! from node 1 to node_count !!!!!!!!!
 for (node=1; node<=node_count; node++)
 {
   if (!P_CheckSight (actor, botnodes[node].mo) )
       continue;
   if (botnodes[node].special2 == 1 && botnodes[node].special1 != 5)
	   continue;
   if (!B_CheckMovePtoP (actor, actor, botnodes[node].mo))
       continue;     // impossible to get to this node
   return node;
 }

 return 0;
}


//#define STOPSPEED               0x1000
#define FRICTION                0xe800   //90.625

//
// B_MoveToNextAltNode
//
void B_MoveToNextAltNode (mobj_t* actor)
{
 int       node;
 Bot_t*    b;

  b = &bots[actor->botnum];

  if (!node_count) return; // no nodes

  node = b->routeposition;

//  Printf (PRINT_HIGH, "%d\n", b->roam);  

  if (b->lastvisitednode && botnodes[b->lastvisitednode].special1 == 5 && actor->reactiontime)
     return; // waiting at the "wait" node

  if (!B_CheckMovePtoP (actor, actor, botnodes[node].mo) )
  {
	  // a bot will roam 200 tics looking for a node
	  if (b->roam == -1)  b->roam = 200;

	  if (b->roam == 0)
	  {        
		if (B_CheckMovePtoP (actor, actor, botnodes[1].mo) )
		{
           node = b->routeposition = 1;
		}
        else
		{
           node = b->routeposition = B_FindAnyVisNode (actor);
		   if (!node)
		   {
			   B_ClearBotRoute(actor);
			   return;
		   }
		}
	  }

	  if (b->roam > 0)
	  {
		  int nextnode;

		  if (b->roam%50 == 0)
			  B_NewChaseDir(actor);

		  nextnode = B_FindNextVisNode (actor);
		  
		  if (!nextnode)
		  {
		    b->roam--;
		    b->forwardmove = 25;		  
		    return;
		  }
		  else
		  {
		    b->routeposition = node = nextnode;
		  }
	  }
  }
  
  b->roam = -1;

  actor->mangle = R_PointToAngle2 (actor->x,
                                   actor->y,
                                  botnodes[node].mo->x,
                                  botnodes[node].mo->y);
  if (!actor->target) actor->angle = actor->mangle;

  if (botnodes[b->lastvisitednode].special1 == 2) //"move slower" node
      b->forwardmove = 25;
  else if (botnodes[node].special1 == 3) // next node is "stop" node
  {
	fixed_t dist;

    dist = P_AproxDistance (botnodes[node].mo->x - actor->x,
                            botnodes[node].mo->y - actor->y );

    if (dist < 17*FRACUNIT) b->forwardmove = 10;
    else                    b->forwardmove = 50;
  }
  else
	b->forwardmove = 50;

  if ( (abs(actor->x - botnodes[node].mo->x) <= 50*FRACUNIT &&
        abs(actor->y - botnodes[node].mo->y) <= 50*FRACUNIT &&
        P_AproxDistance (actor->x + actor->momx - botnodes[node].mo->x, actor->y - botnodes[node].mo->y)
        <= P_AproxDistance (actor->x + actor->momx - botnodes[node].mo->x, actor->y + actor->momy - botnodes[node].mo->y) )
      || (abs(actor->x - botnodes[node].mo->x) <= 2*FRACUNIT &&
          abs(actor->y - botnodes[node].mo->y) <= 2*FRACUNIT)
     )
    {      
	  b->lastvisitednode = node;
      b->routeposition++;

      // end of the route
	  if ( b->routeposition > node_count ||
		   botnodes[node].special2 == 2 )
      {
         if (   P_CheckSight (actor, botnodes[1].mo)
             && B_CheckMovePtoP (actor, actor, botnodes[1].mo) )
		 {		  
		    b->routeposition = 1;
            b->lastvisitednode = 0;
		 }
		 else
		 {
			b->routeposition = B_FindAnyVisNode2(actor);
			b->lastvisitednode = b->routeposition - 1;
		 }
      }

	  if (botnodes[node].special1 == 3) // stop node
	  {
		b->mo->momx /= 6;
	    b->mo->momy /= 6;
	  }
      
	  if (botnodes[node].special1 == 4)    // "Use" node
      {
        b->mo->angle = botnodes[node].mo->angle;
        B_UseLines (b->mo);   // activate a switch
		//use node is always stop node
		b->mo->momx /= 6;
	    b->mo->momy /= 6;
      }

      if (botnodes[node].special1 == 5)    // "wait" node
      {
        actor->reactiontime = botnodes[node].special2;

		//wait node is always stop node
		b->mo->momx /= 6;
	    b->mo->momy /= 6;
      }
	  
	  if (b->destnode == node) 
         B_ClearBotRoute(actor);
    }

}



